from . import lib_etude_courbe as libec
import libargfunc as libargv
import os
import libcurses as curses
from sys import argv,exit
import documentex
from subprocess import call
from pylatex import Document, Section, Itemize, Command, Math, NoEscape
from pylatex.base_classes import Arguments 
from sympy import oo

def etude_courbe():

    if len(argv) != 2:
        print("usage : etude_courbe nom_dossier")
        exit(1)

    dossier = argv[1]

    ok = False
    while not ok:
        x = curses.saisie_message_std('x(t)= ? ')
        x, ok = libargv.sympify(x)
        if not ok:
            curses.afficher_message_std(x)
    # en sortie, expression est une expression sympy utilisable
    ok = False
    while not ok:
        y = curses.saisie_message_std('y(t)= ? ')
        y, ok = libargv.sympify(y)
        if not ok:
            curses.afficher_message_std(y)

    ok = False
    while not ok:
        intervalle_etude = curses.saisie_message_std("intervalle d'étude ? ")
        intervalle_etude, ok = libargv.sympify_intervalle(intervalle_etude)
        if not ok:
            curses.afficher_message_std(intervalle_etude)

    develop = ( curses.saisie_message_std("Voulez-vous également l'étude et le tracé de la développée ? (o/n) ") == 'o' )

    f = libec.courbe(x,y,intervalle_etude)

    xmin=-10
    xmax=10
    ymin=-10
    ymax=10
    if f.i.inf != -oo:
        tmin = round(float(f.i.inf),2)
    else:
        tmin = -10
    if f.i.sup != oo:
        tmax = round(float(f.i.sup),2)
    else:
        tmax = 10

    numero=str(1)
    numerodev=str(int(numero)+1)
    call(['documentex',dossier,'doc'])
    currdir = os.getcwd()
    os.chdir(os.path.join(dossier))
    f.graphe(xmin,xmax,ymin,ymax,tmin,tmax,n=1000)
    if develop:
        dev = f.developpee()
        libec.graphe([f,dev],xmin,xmax,ymin,ymax,tmin,tmax,nom_pdf='courbe_dev.pdf',show=False)
    os.chdir(currdir)


    doc = Section(NoEscape(r"\'Etude d'une courbe"))
    doc.append("Nous étudions la courbe donnée par")
    doc.append(f.latex_print())
    doc.append("Les fonctions de coordonnées sont dérivables et données par")
    doc.append(f.df().latex_print("'"))
    tableau_variation = f.tikztabvar()
    if tableau_variation:
        doc.append("On en déduit le tableau de variation suivant")
        doc.append(tableau_variation)
        etude,var = f.points_particuliers()
        if var:
            doc.append("De ce tableau de variations on déduit les informations suivantes")
            doc.append(etude)
    else:
        doc.append("<+sympy.solve et sympy.solveset sont malheureusement incapables de trouver les points critiques des fonctions de coordonnées pour une raison qui m'échappe+>")
    doc.append(NoEscape(r"Le graphique de la courbe est donné sur la figure \ref{fig:"+numero+"}"))
    doc.append(f.display_graphe(r'Graphique de la courbe',numero))
    if develop:
        doc.append(r'La courbure de la courbe est donnée par')
        with doc.create(Math(escape=False)) as courbure:
            courbure.append(libec.latex(f.courbure()))
        doc.append(r'Sa développée est alors donnée par')
        doc.append(dev.latex_print("_d"))
        doc.append(r"L'étude de la développée donne alors le tableau de variation suivant")
        tabvardev = dev.tikztabvar(der="_d")
        if tabvardev:
            doc.append(tabvardev)
            etude,var = dev.points_particuliers()
            if var:
                doc.append("On a alors les informations suivantes")
                doc.append(etude)
        else:
            doc.append("<+sympy.solve et sympy.solveset sont malheureusement incapables de trouver les points critiques des fonctions de coordonnées pour une raison qui m'échappe+>")
        doc.append(NoEscape(r"On peut alors représenter sur un même graphique la courbe et sa développée, voir figure \ref{fig:"+numerodev+"}"))
        doc.append(libec.display_graphe([f,dev],r'La courbe et sa développée',numerodev,nom_pdf='courbe_dev.pdf'))
    #    doc.append(display_graphe([f],r'La courbe',numerodev,xmin,xmax,ymin,ymax,tmin,tmax,nom_pdf='courbe_dev.pdf',show=False))
    doc.generate_tex(os.path.join(dossier,'example'))
